import {
  Timestamp,
  TimestampUpgradeResponse,
  TimestampUpgradeVerifyResponse
} from '../types'
import { retry, retryAll } from './retry.ts'
import { bytesToHex } from '@noble/hashes/utils'
import { utf8Encoder } from 'nostr-tools/utils'
import { hexStringToUint8Array } from './string.ts'
import { isPromiseFulfilled } from './utils.ts'

/**
 * Generates a timestamp for the provided nostr event ID.
 * @returns Timestamp with its value and the nostr event ID.
 */
export const generateTimestamp = async (
  nostrId: string
): Promise<Timestamp | undefined> => {
  try {
    return {
      timestamp: await retry(() => timestamp(nostrId)),
      nostrId: nostrId
    }
  } catch (error) {
    console.error(error)
    return
  }
}

export const upgradeAndVerifyTimestamp = async (
  timestamp: Timestamp
): Promise<TimestampUpgradeVerifyResponse> => {
  const upgradedResult = await retry(() => upgrade(timestamp))
  const verifiedResult = await verify(upgradedResult)

  console.log('verifiedResult: ', verifiedResult)

  return verifiedResult
}

export const upgradeTimestamps = async (
  timestamps: Timestamp[]
): Promise<TimestampUpgradeResponse[]> => {
  const resolved = await retryAll(timestamps.map((t) => () => upgrade(t)))
  return resolved.filter(isPromiseFulfilled).map((res) => res.value)
}

export const verifyTimestamps = async (timestamps: Timestamp[]) => {
  const settledResults = await Promise.allSettled(
    timestamps.map(async (timestamp) => verify(timestamp))
  )
  return settledResults.filter(isPromiseFulfilled).map((res) => res.value)
}

const upgrade = async (
  t: Timestamp
): Promise<TimestampUpgradeVerifyResponse> => {
  let detachedTimestamp =
    window.OpenTimestamps.DetachedTimestampFile.deserialize(
      hexStringToUint8Array(t.timestamp)
    )

  const changed: boolean =
    await window.OpenTimestamps.upgrade(detachedTimestamp)
  if (changed) {
    const updated = detachedTimestamp.serializeToBytes()

    const value = {
      ...t,
      timestamp: bytesToHex(updated)
    }

    return {
      value,
      upgraded: true,
      isComplete: detachedTimestamp.timestamp.isTimestampComplete()
    }
  }
  return {
    value: t,
    upgraded: false,
    isComplete: detachedTimestamp.timestamp.isTimestampComplete()
  }
}

export const verify = async (
  t: TimestampUpgradeVerifyResponse
): Promise<TimestampUpgradeVerifyResponse> => {
  const detachedNostrId = window.OpenTimestamps.DetachedTimestampFile.fromBytes(
    new window.OpenTimestamps.Ops.OpSHA256(),
    utf8Encoder.encode(t.value.nostrId)
  )

  let detachedTimestamp =
    window.OpenTimestamps.DetachedTimestampFile.deserialize(
      hexStringToUint8Array(t.value.timestamp)
    )

  const res = await window.OpenTimestamps.verify(
    detachedTimestamp,
    detachedNostrId
  )

  return {
    ...t,
    verified: !!res.bitcoin,
    verification: res?.bitcoin || null
  }
}

const timestamp = async (nostrId: string): Promise<string> => {
  const detachedTimestamp =
    window.OpenTimestamps.DetachedTimestampFile.fromBytes(
      new window.OpenTimestamps.Ops.OpSHA256(),
      utf8Encoder.encode(nostrId)
    )

  await window.OpenTimestamps.stamp(detachedTimestamp)
  const ctx = new window.OpenTimestamps.Context.StreamSerialization()
  detachedTimestamp.serialize(ctx)
  const timestampBytes = ctx.getOutput()
  return bytesToHex(timestampBytes)
}
